;; Generated package description from sotlisp.el  -*- no-byte-compile: t -*-
(define-package "sotlisp" "1.6.2.0.20190211.202614" "Write lisp at the speed of thought." '((emacs "24.1")) :keywords '("convenience" "lisp") :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :url "https://github.com/Malabarba/speed-of-thought-lisp")
